(* Copyright (C) 1995, Digital Equipment Corporation.       *)
(* All rights reserved.                                     *)
(* See the file COPYRIGHT for a full description.           *)
(*                                                          *)
(* Last modified on Thu Feb  9 08:56:02 PST 1995 by kalsow  *)
(*      modified on Mon Jun  6 03:18:55 PDT 1994 by mhb   *)
(*      modified on Tue Feb 16 16:31:40 PST 1993 by johnh *)

(*********************************************************************
|*  NOTE: This file is generated automatically from the event
|*        definition file DGraph.evt.
|*********************************************************************)

<* PRAGMA LL *>

MODULE DGraphTranscriptView;

<*NOWARN*> IMPORT AdjMatrix, TextPort, Rd, ZeusClass;
<*NOWARN*> IMPORT DGraphViewClass, Filter, TextEditVBT, Fmt, ZFmt;
<*NOWARN*> IMPORT Wr, ZeusPanel, FormsVBT, VBT, View;

<* FATAL FormsVBT.Error, FormsVBT.Unimplemented *>

REVEAL
  T = Public BRANDED OBJECT
    fv: FormsVBT.T := NIL;
    te: TextEditVBT.T := NIL;
  OVERRIDES
    init       := TViewInit;
    install    := TViewInstall;
    delete     := TViewDelete;
    snapshot   := TViewSnapshot;
    restore    := TViewRestore;
    config     := TViewConfig;
    reactivity := TViewReactivity;
    startrun   := TViewStartrun;
    endrun     := TViewEndrun;
    oeSetup := Setup;
    oeAddEdge := AddEdge;
    oeMarkEdge := MarkEdge;
    oeUnMarkEdge := UnMarkEdge;
    oeMarkVertex := MarkVertex;
    oeUnMarkVertex := UnMarkVertex;
    oeNewTree := NewTree;
    oeHighlight := Highlight;
    oeAddChild := AddChild;
    oeRemoveChild := RemoveChild;
  END;

PROCEDURE TViewInit (view: T): T =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "init");
    RETURN DGraphViewClass.T.init (view, NIL);
  END TViewInit;

PROCEDURE Clear (<* UNUSED *> fv  : FormsVBT.T;
                 <* UNUSED *> name: TEXT;
                              cl  : REFANY;
                 <* UNUSED *> time: VBT.TimeStamp) =
  BEGIN
    TextPort.SetText(NARROW(cl, T).te.tp, "")
  END Clear;

PROCEDURE TViewInstall (view: T) =
  <* LL = VBT.mu *>
  BEGIN
    view.fv := ZeusPanel.NewForm("DGraphTranscriptView.fv");
    view.te := FormsVBT.GetVBT(view.fv, "transcript");
    TViewZTrace (view, "install");
    FormsVBT.AttachProc(view.fv, "clear", Clear, view);
    EVAL Filter.Replace (view, view.fv);
    DGraphViewClass.T.install (view);
  END TViewInstall;

PROCEDURE TViewDelete (view: T) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "delete");
    DGraphViewClass.T.delete (view);
   END TViewDelete;

PROCEDURE TViewSnapshot (view: T; wr: Wr.T) RAISES {ZeusClass.Error} =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "snapshot");
    DGraphViewClass.T.snapshot (view, wr);
   END TViewSnapshot;

PROCEDURE TViewRestore (view: T; rd: Rd.T) RAISES {ZeusClass.Error} =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "restore");
    DGraphViewClass.T.restore (view, rd);
   END TViewRestore;

PROCEDURE TViewConfig (
    view: T;
    state: ZeusClass.StateChange;
    o: ZeusClass.T) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "config");
    DGraphViewClass.T.config (view, state, o);
   END TViewConfig;

PROCEDURE TViewReactivity (view: T; <*UNUSED*> on: BOOLEAN) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace(view, "reactivity");
    DGraphViewClass.T.reactivity (view, TRUE);
  END TViewReactivity;

PROCEDURE TViewStartrun (view: T) =
  <* LL = {} *>
  BEGIN
    TViewZTrace (view, "startrun");
    DGraphViewClass.T.startrun (view);
   END TViewStartrun;

PROCEDURE TViewEndrun (view: T) =
  <* LL = {} *>
  BEGIN
    TViewZTrace (view, "endrun");
    DGraphViewClass.T.endrun (view);
  END TViewEndrun;


(* event handling methods: *)

PROCEDURE Setup (view: T;  m: AdjMatrix.T) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Setup ...")
       ELSE
          TViewTrace (view, "Setup "
          & AdjMatrix.ToText(m)
          )
       END
    END
    END
  END Setup;

PROCEDURE AddEdge (view: T;  from, to: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "AddEdge ...")
       ELSE
          TViewTrace (view, "AddEdge "
          & Fmt.Int(from)
          & " "
          & Fmt.Int(to)
          )
       END
    END
    END
  END AddEdge;

PROCEDURE MarkEdge (view: T;  from, to: INTEGER; depth: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "MarkEdge ...")
       ELSE
          TViewTrace (view, "MarkEdge "
          & Fmt.Int(from)
          & " "
          & Fmt.Int(to)
          & " "
          & Fmt.Int(depth)
          )
       END
    END
    END
  END MarkEdge;

PROCEDURE UnMarkEdge (view: T;  from, to: INTEGER; depth: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "UnMarkEdge ...")
       ELSE
          TViewTrace (view, "UnMarkEdge "
          & Fmt.Int(from)
          & " "
          & Fmt.Int(to)
          & " "
          & Fmt.Int(depth)
          )
       END
    END
    END
  END UnMarkEdge;

PROCEDURE MarkVertex (view: T;  v: INTEGER; depth: INTEGER; rcset: AdjMatrix.RCSet) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "MarkVertex ...")
       ELSE
          TViewTrace (view, "MarkVertex "
          & Fmt.Int(v)
          & " "
          & Fmt.Int(depth)
          & " "
          & AdjMatrix.RCToText(rcset)
          )
       END
    END
    END
  END MarkVertex;

PROCEDURE UnMarkVertex (view: T;  v: INTEGER; depth: INTEGER; rcset: AdjMatrix.RCSet) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "UnMarkVertex ...")
       ELSE
          TViewTrace (view, "UnMarkVertex "
          & Fmt.Int(v)
          & " "
          & Fmt.Int(depth)
          & " "
          & AdjMatrix.RCToText(rcset)
          )
       END
    END
    END
  END UnMarkVertex;

PROCEDURE NewTree (view: T;  root: INTEGER; label: TEXT) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "NewTree ...")
       ELSE
          TViewTrace (view, "NewTree "
          & Fmt.Int(root)
          & " "
          & (label)
          )
       END
    END
    END
  END NewTree;

PROCEDURE Highlight (view: T;  node: INTEGER; highlight: REAL; nodeOnly: BOOLEAN) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Highlight ...")
       ELSE
          TViewTrace (view, "Highlight "
          & Fmt.Int(node)
          & " "
          & Fmt.Real(highlight)
          & " "
          & Fmt.Bool(nodeOnly)
          )
       END
    END
    END
  END Highlight;

PROCEDURE AddChild (view: T;  parent, pred, child: INTEGER; label: TEXT) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "AddChild ...")
       ELSE
          TViewTrace (view, "AddChild "
          & Fmt.Int(parent)
          & " "
          & Fmt.Int(pred)
          & " "
          & Fmt.Int(child)
          & " "
          & (label)
          )
       END
    END
    END
  END AddChild;

PROCEDURE RemoveChild (view: T;  parent, ch: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "RemoveChild ...")
       ELSE
          TViewTrace (view, "RemoveChild "
          & Fmt.Int(parent)
          & " "
          & Fmt.Int(ch)
          )
       END
    END
    END
  END RemoveChild;





PROCEDURE TViewZTrace (view: T; t: TEXT) =
  BEGIN
    IF view.fv # NIL THEN
      IF FormsVBT.GetBoolean(view.fv, "zeus") THEN
        TextPort.PutText(view.te.tp, "**zeus:  " & t & "\n");
        TextPort.Normalize(view.te.tp, LAST(INTEGER))
      END
    END
  END TViewZTrace;

PROCEDURE TViewTrace (view: T; t: TEXT) =
  BEGIN
    TextPort.PutText(view.te.tp, "--event: " & t & "\n");
    TextPort.Normalize(view.te.tp, LAST(INTEGER))
  END TViewTrace;


PROCEDURE TViewNew (): View.T =
  BEGIN
    RETURN NEW(T).init()
  END TViewNew;


BEGIN
  ZeusPanel.RegisterView
      (TViewNew, "DGraph Transcript View", "DGraph");
END DGraphTranscriptView.
