(* Copyright (C) 1995, Digital Equipment Corporation.       *)
(* All rights reserved.                                     *)
(* See the file COPYRIGHT for a full description.           *)
(*                                                          *)
(* Last modified on Thu Feb  9 08:56:02 PST 1995 by kalsow  *)
(*      modified on Mon Jun  6 03:18:55 PDT 1994 by mhb   *)
(*      modified on Tue Feb 16 16:31:40 PST 1993 by johnh *)

(*********************************************************************
|*  NOTE: This file is generated automatically from the event
|*        definition file PQueue.evt.
|*********************************************************************)

<* PRAGMA LL *>

MODULE PQueueTranscriptView;

<*NOWARN*> IMPORT TextPort, PQueue, Rd, ZeusClass, Filter;
<*NOWARN*> IMPORT TextEditVBT, Fmt, ZFmt, Wr, ZeusPanel, FormsVBT;
<*NOWARN*> IMPORT VBT, View, PQueueViewClass;

<* FATAL FormsVBT.Error, FormsVBT.Unimplemented *>

REVEAL
  T = Public BRANDED OBJECT
    fv: FormsVBT.T := NIL;
    te: TextEditVBT.T := NIL;
  OVERRIDES
    init       := TViewInit;
    install    := TViewInstall;
    delete     := TViewDelete;
    snapshot   := TViewSnapshot;
    restore    := TViewRestore;
    config     := TViewConfig;
    reactivity := TViewReactivity;
    startrun   := TViewStartrun;
    endrun     := TViewEndrun;
    oeSetup := Setup;
    oeInitSort := InitSort;
    oeInsert := Insert;
    oeHeapOpInit := HeapOpInit;
    oeUpHeapStep := UpHeapStep;
    oeHeapStep := HeapStep;
    oePlaceElement := PlaceElement;
    oeRemove := Remove;
    oeCompare := Compare;
    oePause := Pause;
    oeSortStep := SortStep;
  END;

PROCEDURE TViewInit (view: T): T =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "init");
    RETURN PQueueViewClass.T.init (view, NIL);
  END TViewInit;

PROCEDURE Clear (<* UNUSED *> fv  : FormsVBT.T;
                 <* UNUSED *> name: TEXT;
                              cl  : REFANY;
                 <* UNUSED *> time: VBT.TimeStamp) =
  BEGIN
    TextPort.SetText(NARROW(cl, T).te.tp, "")
  END Clear;

PROCEDURE TViewInstall (view: T) =
  <* LL = VBT.mu *>
  BEGIN
    view.fv := ZeusPanel.NewForm("PQueueTranscriptView.fv");
    view.te := FormsVBT.GetVBT(view.fv, "transcript");
    TViewZTrace (view, "install");
    FormsVBT.AttachProc(view.fv, "clear", Clear, view);
    EVAL Filter.Replace (view, view.fv);
    PQueueViewClass.T.install (view);
  END TViewInstall;

PROCEDURE TViewDelete (view: T) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "delete");
    PQueueViewClass.T.delete (view);
   END TViewDelete;

PROCEDURE TViewSnapshot (view: T; wr: Wr.T) RAISES {ZeusClass.Error} =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "snapshot");
    PQueueViewClass.T.snapshot (view, wr);
   END TViewSnapshot;

PROCEDURE TViewRestore (view: T; rd: Rd.T) RAISES {ZeusClass.Error} =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "restore");
    PQueueViewClass.T.restore (view, rd);
   END TViewRestore;

PROCEDURE TViewConfig (
    view: T;
    state: ZeusClass.StateChange;
    o: ZeusClass.T) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace (view, "config");
    PQueueViewClass.T.config (view, state, o);
   END TViewConfig;

PROCEDURE TViewReactivity (view: T; <*UNUSED*> on: BOOLEAN) =
  <* LL = VBT.mu *>
  BEGIN
    TViewZTrace(view, "reactivity");
    PQueueViewClass.T.reactivity (view, TRUE);
  END TViewReactivity;

PROCEDURE TViewStartrun (view: T) =
  <* LL = {} *>
  BEGIN
    TViewZTrace (view, "startrun");
    PQueueViewClass.T.startrun (view);
   END TViewStartrun;

PROCEDURE TViewEndrun (view: T) =
  <* LL = {} *>
  BEGIN
    TViewZTrace (view, "endrun");
    PQueueViewClass.T.endrun (view);
  END TViewEndrun;


(* event handling methods: *)

PROCEDURE Setup (view: T;  size: INTEGER; doSort: BOOLEAN) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Setup ...")
       ELSE
          TViewTrace (view, "Setup "
          & Fmt.Int(size)
          & " "
          & Fmt.Bool(doSort)
          )
       END
    END
    END
  END Setup;

PROCEDURE InitSort (view: T;  vals: PQueue.Array) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "InitSort ...")
       ELSE
          TViewTrace (view, "InitSort "
          & PQueue.PrintHeap(vals)
          )
       END
    END
    END
  END InitSort;

PROCEDURE Insert (view: T;  el: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Insert ...")
       ELSE
          TViewTrace (view, "Insert "
          & Fmt.Int(el)
          )
       END
    END
    END
  END Insert;

PROCEDURE HeapOpInit (view: T;  k: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "HeapOpInit ...")
       ELSE
          TViewTrace (view, "HeapOpInit "
          & Fmt.Int(k)
          )
       END
    END
    END
  END HeapOpInit;

PROCEDURE UpHeapStep (view: T;  k: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "UpHeapStep ...")
       ELSE
          TViewTrace (view, "UpHeapStep "
          & Fmt.Int(k)
          )
       END
    END
    END
  END UpHeapStep;

PROCEDURE HeapStep (view: T;  k, n: INTEGER; down: BOOLEAN) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "HeapStep ...")
       ELSE
          TViewTrace (view, "HeapStep "
          & Fmt.Int(k)
          & " "
          & Fmt.Int(n)
          & " "
          & Fmt.Bool(down)
          )
       END
    END
    END
  END HeapStep;

PROCEDURE PlaceElement (view: T;  k: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "PlaceElement ...")
       ELSE
          TViewTrace (view, "PlaceElement "
          & Fmt.Int(k)
          )
       END
    END
    END
  END PlaceElement;

PROCEDURE Remove (view: T;  ) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Remove ...")
       ELSE
          TViewTrace (view, "Remove "
          )
       END
    END
    END
  END Remove;

PROCEDURE Compare (view: T;  k: INTEGER; n: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Compare ...")
       ELSE
          TViewTrace (view, "Compare "
          & Fmt.Int(k)
          & " "
          & Fmt.Int(n)
          )
       END
    END
    END
  END Compare;

PROCEDURE Pause (view: T;  ) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "Pause ...")
       ELSE
          TViewTrace (view, "Pause "
          )
       END
    END
    END
  END Pause;

PROCEDURE SortStep (view: T;  k: INTEGER) =
  <* LL = {} *>
  BEGIN
    LOCK VBT.mu DO
    IF FormsVBT.GetBoolean(view.fv, "alg") THEN
       IF NOT FormsVBT.GetBoolean(view.fv, "args") THEN
          TViewTrace (view, "SortStep ...")
       ELSE
          TViewTrace (view, "SortStep "
          & Fmt.Int(k)
          )
       END
    END
    END
  END SortStep;





PROCEDURE TViewZTrace (view: T; t: TEXT) =
  BEGIN
    IF view.fv # NIL THEN
      IF FormsVBT.GetBoolean(view.fv, "zeus") THEN
        TextPort.PutText(view.te.tp, "**zeus:  " & t & "\n");
        TextPort.Normalize(view.te.tp, LAST(INTEGER))
      END
    END
  END TViewZTrace;

PROCEDURE TViewTrace (view: T; t: TEXT) =
  BEGIN
    TextPort.PutText(view.te.tp, "--event: " & t & "\n");
    TextPort.Normalize(view.te.tp, LAST(INTEGER))
  END TViewTrace;


PROCEDURE TViewNew (): View.T =
  BEGIN
    RETURN NEW(T).init()
  END TViewNew;


BEGIN
  ZeusPanel.RegisterView
      (TViewNew, "PQueue Transcript View", "PQueue");
END PQueueTranscriptView.
