(* Copyright 1996-1998 John D. Polstra.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by John D. Polstra.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: RCS_Date.i3,v 1.1.1.1 2010-08-07 16:18:32 wagner Exp $ *)

INTERFACE RCS_Date;

IMPORT Time;

EXCEPTION Error(TEXT);

TYPE
  T = TEXT;

PROCEDURE Compare(a, b: T): [-1..1];
(* Compares two RCS dates. *)

PROCEDURE Equal(a, b: T): BOOLEAN;
(* Returns an indication of whether two RCS dates are equal. *)

PROCEDURE FromTime(t: Time.T): T;

PROCEDURE ToTime(d: T): Time.T
  RAISES {Error};

PROCEDURE ToTimeApprox(d: T): Time.T;
  (* same as to time, but does not raise any exceptions (issues error
     messages to stderr instead) and tries to correct impossible
     values *)

PROCEDURE Valid(d: T): BOOLEAN;
(* Checks the date for validity. *)

END RCS_Date.
